"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var interfaces_1 = require("./interfaces");
var type_formatter_1 = require("./type-formatter");
var assign_1 = __importDefault(require("lodash/assign"));
var defaults_1 = __importDefault(require("lodash/defaults"));
var mapKeys_1 = __importDefault(require("lodash/mapKeys"));
var pick_1 = __importDefault(require("lodash/pick"));
var keys_1 = __importDefault(require("lodash/keys"));
var cloneDeep_1 = __importDefault(require("lodash/cloneDeep"));
var isFunction_1 = __importDefault(require("lodash/isFunction"));
var forEach_1 = __importDefault(require("lodash/forEach"));
var set_1 = __importDefault(require("lodash/set"));
var lodash_1 = require("lodash");
var utils_1 = require("./utils");
/**
 * Generate Swagger schema content describing the specified Sails attribute.
 *
 * XXX TODO: add test to this function
 *
 * @see https://swagger.io/docs/specification/data-models/
 * @param {Record<string, any>} attribute Sails model attribute specification as per `Model.js` file
 */
exports.generateAttributeSchema = function (attribute) {
    var _a, _b, _c, _d, _e, _f;
    var ai = attribute || {}, sts = type_formatter_1.swaggerTypes;
    var type = ai.type || 'string';
    var columnType = (_a = ai.autoMigrations) === null || _a === void 0 ? void 0 : _a.columnType;
    var autoIncrement = (_b = ai.autoMigrations) === null || _b === void 0 ? void 0 : _b.autoIncrement;
    var schema = {};
    var formatDesc = function (extra) {
        var ret = [];
        if (ai.description)
            ret.push(ai.description);
        if (extra)
            ret.push(extra);
        return ret.join(' ');
    };
    if (((_c = ai.meta) === null || _c === void 0 ? void 0 : _c.swagger) && 'type' in ai.meta.swagger) {
        // OpenAPI 3 stipulates NO type as 'any', allow this by 'type' present but null to achieve this
        if (ai.meta.swagger.type)
            schema.type = ai.meta.swagger.type;
    }
    else if (ai.model) {
        assign_1.default(schema, {
            description: formatDesc("JSON dictionary representing the **" + ai.model + "** instance or FK when creating / updating / not populated"),
            // '$ref': '#/components/schemas/' + ai.model,
            oneOf: [
                { '$ref': '#/components/schemas/' + ai.model },
            ],
        });
    }
    else if (ai.collection) {
        assign_1.default(schema, {
            description: formatDesc("Array of **" + ai.collection + "**'s or array of FK's when creating / updating / not populated"),
            type: 'array',
            items: { '$ref': '#/components/schemas/' + ai.collection },
        });
    }
    else if (type == 'number') {
        var t = autoIncrement ? sts.integer : sts.double;
        if ((_d = ai.validations) === null || _d === void 0 ? void 0 : _d.isInteger) {
            t = sts.integer;
        }
        else if (columnType) {
            var ct = columnType;
            if (ct.match(/int/i))
                t = sts.integer;
            else if (ct.match(/long/i))
                t = sts.long;
            else if (ct.match(/float/i))
                t = sts.float;
            else if (ct.match(/double/i))
                t = sts.double;
            else if (ct.match(/decimal/i))
                t = sts.double;
        }
        assign_1.default(schema, t);
    }
    else if (type == 'boolean') {
        assign_1.default(schema, sts.boolean);
    }
    else if (type == 'json') {
        assign_1.default(schema, sts.string);
    }
    else if (type == 'ref') {
        var t = sts.string;
        if (columnType) {
            var ct = columnType;
            if (ct.match(/timestamp/i))
                t = sts.datetime;
            else if (ct.match(/datetime/i))
                t = sts.datetime;
            else if (ct.match(/date/i))
                t = sts.date;
        }
        assign_1.default(schema, t);
    }
    else { // includes =='string'
        assign_1.default(schema, sts.string);
    }
    var isIP = false;
    if (schema.type == 'string') {
        var v = ai.validations;
        if (v) {
            if (v.isEmail)
                schema.format = 'email';
            if (v.isIP) {
                isIP = true;
                schema.format = 'ipv4';
            }
            if (v.isURL)
                schema.format = 'uri';
            if (v.isUUID)
                schema.format = 'uuid';
            if (v.regex)
                schema.pattern = v.regex.toString().slice(1, -1);
        }
    }
    var annotations = [];
    // annotate format with Sails autoCreatedAt/autoUpdatedAt
    if (schema.type == 'string' && schema.format == 'date-time') {
        if (ai.autoCreatedAt)
            annotations.push('autoCreatedAt');
        else if (ai.autoUpdatedAt)
            annotations.push('autoUpdatedAt');
    }
    // process Sails --> Swagger attribute mappings as per sailAttributePropertiesMap
    defaults_1.default(schema, mapKeys_1.default(pick_1.default(ai, keys_1.default(type_formatter_1.sailsAttributePropertiesMap)), function (v, k) { return type_formatter_1.sailsAttributePropertiesMap[k]; }));
    // process Sails --> Swagger attribute mappings as per validationsMap
    defaults_1.default(schema, mapKeys_1.default(pick_1.default(ai.validations, keys_1.default(type_formatter_1.validationsMap)), function (v, k) { return type_formatter_1.validationsMap[k]; }));
    // copy default into example if present
    if (schema.default && !schema.example) {
        schema.example = schema.default;
    }
    // process final autoMigrations: autoIncrement, unique
    if (autoIncrement) {
        annotations.push('autoIncrement');
    }
    if ((_e = ai.autoMigrations) === null || _e === void 0 ? void 0 : _e.unique) {
        schema.uniqueItems = true;
    }
    // represent Sails `isIP` as one of ipv4/ipv6
    if (schema.type == 'string' && isIP) {
        schema = {
            description: formatDesc('ipv4 or ipv6 address'),
            oneOf: [
                cloneDeep_1.default(schema),
                assign_1.default(cloneDeep_1.default(schema), { format: 'ipv6' }),
            ]
        };
    }
    if (annotations.length > 0) {
        var s = "Note Sails special attributes: " + annotations.join(', ');
        schema.description = schema.description ? schema.description + "\n\n" + s : s;
    }
    if (schema.description)
        schema.description = schema.description.trim();
    // note: required --> required[] (not here, needs to be done at model level)
    // finally, overwrite in custom swagger
    if ((_f = ai.meta) === null || _f === void 0 ? void 0 : _f.swagger) {
        // note: 'type' handled above
        assign_1.default(schema, lodash_1.omit(ai.meta.swagger, 'exclude', 'type'));
    }
    return schema;
};
/**
 * Generate the OpenAPI schemas for the foreign key values used to reference
 * ORM records for the associations of the specified Sails Model.
 *
 * Used for 'replace' REST blueprint.
 *
 * @param model
 * @param models
 */
exports.generateModelAssociationFKAttributeSchemas = function (model, aliasesToInclude, models) {
    if (!model.associations) {
        return [];
    }
    return model.associations.map(function (association) {
        if (!aliasesToInclude || aliasesToInclude.indexOf(association.alias) < 0)
            return;
        var targetModelIdentity = association.type === 'model' ? association.model : association.collection;
        var targetModel = models[targetModelIdentity];
        if (!targetModel) {
            return; // data structure integrity issue should not occur
        }
        var description = association.type === 'model' ?
            "**" + model.globalId + "** record's foreign key value to use as the replacement for this attribute"
            : "**" + model.globalId + "** record's foreign key values to use as the replacement for this collection";
        var targetFKAttribute = targetModel.attributes[targetModel.primaryKey];
        return exports.generateAttributeSchema(__assign(__assign({}, targetFKAttribute), { autoMigrations: __assign(__assign({}, (targetFKAttribute.autoMigrations || {})), { autoIncrement: false }), description: description + " (**" + association.alias + "** association" + (targetFKAttribute.description ? '; ' + targetFKAttribute.description : '') + ")" }));
    })
        .filter(function (parameter) { return parameter; });
};
/**
 * Generate the OpenAPI parameters for the foreign key values used to reference
 * ORM records for the associations of the specified Sails Model.
 *
 * Used for 'replace' shortcut blueprint.
 *
 * @param model
 * @param aliasesToInclude
 * @param models
 */
exports.generateModelAssociationFKAttributeParameters = function (model, aliasesToInclude, models) {
    if (!model.associations) {
        return [];
    }
    return model.associations.map(function (association) {
        if (!aliasesToInclude || aliasesToInclude.indexOf(association.alias) < 0)
            return;
        var targetModelIdentity = association.type === 'model' ? association.model : association.collection;
        var targetModel = models[targetModelIdentity];
        if (!targetModel) {
            return; // data structure integrity issue should not occur
        }
        var description = association.type === 'model' ?
            "**" + model.globalId + "** record's foreign key value to use as the replacement for this attribute"
            : "**" + model.globalId + "** record's foreign key values to use as the replacement for this collection";
        var targetFKAttribute = targetModel.attributes[targetModel.primaryKey];
        var targetFKAttributeSchema = exports.generateAttributeSchema(__assign(__assign({}, targetFKAttribute), { autoMigrations: __assign(__assign({}, (targetFKAttribute.autoMigrations || {})), { autoIncrement: false }), description: description + " (**" + association.alias + "** association" + (targetFKAttribute.description ? '; ' + targetFKAttribute.description : '') + ")" }));
        return {
            in: 'query',
            name: association.alias,
            description: targetFKAttributeSchema.description,
            schema: {
                type: 'array',
                items: targetFKAttributeSchema,
            },
        };
    })
        .filter(function (parameter) { return parameter; });
};
exports.generateSchemaAsQueryParameters = function (schema) {
    var required = schema.required || [];
    return lodash_1.map(schema.properties || {}, function (property, name) {
        var parameter = {
            in: 'query',
            name: name,
            schema: property,
        };
        if (property.description) {
            parameter.description = property.description;
        }
        if (required.indexOf(name) >= 0) {
            parameter.required = true;
        }
        return parameter;
    });
};
/**
 * Generate Swagger schema content describing specified Sails models.
 *
 * @see https://swagger.io/docs/specification/data-models/
 *
 * @param models parsed Sails models as per `parsers.parseModels()`
 * @returns
 */
exports.generateSchemas = function (models) {
    return Object.keys(models)
        .reduce(function (schemas, identity) {
        var _a, _b, _c, _d;
        var model = models[identity];
        if (((_b = (_a = model.swagger) === null || _a === void 0 ? void 0 : _a.modelSchema) === null || _b === void 0 ? void 0 : _b.exclude) === true) {
            return schemas;
        }
        var schema = __assign({ description: ((_c = model.swagger.modelSchema) === null || _c === void 0 ? void 0 : _c.description) || "Sails ORM Model **" + model.globalId + "**", properties: {}, required: [] }, lodash_1.omit(((_d = model.swagger) === null || _d === void 0 ? void 0 : _d.modelSchema) || {}, 'exclude', 'description', 'tags'));
        var attributes = model.attributes || {};
        defaults_1.default(schema.properties, Object.keys(attributes).reduce(function (props, attributeName) {
            var _a, _b;
            var attribute = model.attributes[attributeName];
            if (((_b = (_a = attribute.meta) === null || _a === void 0 ? void 0 : _a.swagger) === null || _b === void 0 ? void 0 : _b.exclude) !== true) {
                props[attributeName] = exports.generateAttributeSchema(attribute);
                if (attribute.required)
                    schema.required.push(attributeName);
            }
            return props;
        }, {}));
        if (schema.required.length <= 0)
            delete schema.required;
        schemas[model.identity] = schema;
        return schemas;
    }, {});
};
/**
 * Generate Swagger schema content describing specified Sails routes/actions.
 *
 * @see https://swagger.io/docs/specification/paths-and-operations/
 *
 * TODO: break down this function into smaller methods and add tests separately
 *
 * @param routes
 * @param templates
 * @param defaultsValues
 * @param models
 */
exports.generatePaths = function (routes, templates, defaultsValues, specification, models) {
    var paths = {};
    var tags = specification.tags;
    var components = specification.components;
    if (!components.parameters) {
        components.parameters = {};
    }
    forEach_1.default(routes, function (route) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m, _o, _p, _q;
        if (((_a = route.swagger) === null || _a === void 0 ? void 0 : _a.exclude) === true) {
            return;
        }
        var pathEntry;
        var isParam = function (inType, name) {
            return !!pathEntry.parameters
                .map(function (parameter) { return utils_1.resolveRef(specification, parameter); })
                .find(function (parameter) { return parameter && 'in' in parameter && parameter.in == inType && parameter.name == name; });
        };
        if (route.middlewareType === interfaces_1.MiddlewareType.BLUEPRINT && route.model) {
            if (((_c = (_b = route.model.swagger) === null || _b === void 0 ? void 0 : _b.modelSchema) === null || _c === void 0 ? void 0 : _c.exclude) === true) {
                return;
            }
            var template_1 = templates[route.action] || {};
            var subst_1 = function (str) { return str ? str.replace('{globalId}', route.model.globalId) : ''; };
            // handle special case of PK parameter
            var parameters = __spreadArrays((template_1.parameters || [])).map(function (parameter) {
                if (parameter === 'primaryKeyPathParameter') {
                    var primaryKey = route.model.primaryKey;
                    var attributeInfo = route.model.attributes[primaryKey];
                    var pname = 'ModelPKParam-' + route.model.identity;
                    if (components.parameters && !components.parameters[pname]) {
                        components.parameters[pname] = {
                            in: 'path',
                            name: '_' + primaryKey,
                            required: true,
                            schema: exports.generateAttributeSchema(attributeInfo),
                            description: subst_1('The desired **{globalId}** record\'s primary key value'),
                        };
                    }
                    return { $ref: '#/components/parameters/' + pname };
                }
                return parameter;
            });
            pathEntry = __assign(__assign({ summary: subst_1(template_1.summary), description: subst_1(template_1.description), externalDocs: template_1.externalDocs || undefined, tags: ((_d = route.swagger) === null || _d === void 0 ? void 0 : _d.tags) || ((_e = route.model.swagger.modelSchema) === null || _e === void 0 ? void 0 : _e.tags) || ((_g = (_f = route.model.swagger.actions) === null || _f === void 0 ? void 0 : _f.allactions) === null || _g === void 0 ? void 0 : _g.tags) || [route.model.globalId], parameters: parameters, responses: cloneDeep_1.default(defaultsValues.responses || {}) }, cloneDeep_1.default(lodash_1.omit(((_h = route.model.swagger.actions) === null || _h === void 0 ? void 0 : _h.allactions) || {}, 'exclude'))), cloneDeep_1.default(lodash_1.omit(route.swagger || {}, 'exclude')));
            var modifiers_1 = {
                addPopulateQueryParam: function () {
                    var _a;
                    var assoc = ((_a = route.model) === null || _a === void 0 ? void 0 : _a.associations) || [];
                    if (isParam('query', 'populate') || assoc.length == 0)
                        return;
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'populate',
                        required: false,
                        schema: {
                            type: 'string',
                            example: __spreadArrays(['false'], (assoc.map(function (row) { return row.alias; }) || [])).join(','),
                        },
                        description: 'If specified, overide the default automatic population process.'
                            + ' Accepts a comma-separated list of attribute names for which to populate record values,'
                            + ' or specify `false` to have no attributes populated.',
                    });
                },
                addSelectQueryParam: function () {
                    if (isParam('query', 'select'))
                        return;
                    var attributes = route.model.attributes || {};
                    var csv = lodash_1.reduce(attributes, function (acc, a, n) { var _a, _b; return ((((_b = (_a = a.meta) === null || _a === void 0 ? void 0 : _a.swagger) === null || _b === void 0 ? void 0 : _b.exclude) === true) ? acc : __spreadArrays(acc, [n])); }, []);
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'select',
                        required: false,
                        schema: {
                            type: 'string',
                            example: csv.join(','),
                        },
                        description: 'The attributes to include in the result, specified as a comma-delimited list.'
                            + ' By default, all attributes are selected.'
                            + ' Not valid for plural (“collection”) association attributes.',
                    });
                },
                addOmitQueryParam: function () {
                    if (isParam('query', 'omit'))
                        return;
                    var attributes = route.model.attributes || {};
                    var csv = lodash_1.reduce(attributes, function (acc, a, n) { var _a, _b; return ((((_b = (_a = a.meta) === null || _a === void 0 ? void 0 : _a.swagger) === null || _b === void 0 ? void 0 : _b.exclude) === true) ? acc : __spreadArrays(acc, [n])); }, []);
                    pathEntry.parameters.push({
                        in: 'query',
                        name: 'omit',
                        required: false,
                        schema: {
                            type: 'string',
                            example: csv.join(','),
                        },
                        description: 'The attributes to exclude from the result, specified as a comma-delimited list.'
                            + ' Cannot be used in conjuction with `select`.'
                            + ' Not valid for plural (“collection”) association attributes.',
                    });
                },
                addModelBodyParam: function () {
                    var _a;
                    if (route.actionType === 'shortcutBlueprint') {
                        var schema = (_a = specification.components.schemas) === null || _a === void 0 ? void 0 : _a[route.model.identity];
                        if (schema) {
                            var resolvedSchema = utils_1.resolveRef(specification, schema);
                            if (resolvedSchema) {
                                exports.generateSchemaAsQueryParameters(resolvedSchema).map(function (p) {
                                    if (isParam('query', p.name))
                                        return;
                                    pathEntry.parameters.push(p);
                                });
                            }
                        }
                    }
                    else {
                        if (pathEntry.requestBody)
                            return;
                        pathEntry.requestBody = {
                            description: subst_1('JSON dictionary representing the {globalId} instance to create.'),
                            required: true,
                            content: {
                                'application/json': {
                                    schema: { "$ref": "#/components/schemas/" + route.model.identity }
                                },
                            },
                        };
                    }
                },
                addResultOfArrayOfModels: function () {
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: {
                                    type: 'array',
                                    items: { '$ref': '#/components/schemas/' + route.model.identity },
                                },
                            },
                        },
                    });
                },
                addAssociationPathParam: function () {
                    if (isParam('path', 'association'))
                        return;
                    pathEntry.parameters.splice(1, 0, {
                        in: 'path',
                        name: 'association',
                        required: true,
                        schema: {
                            type: 'string',
                            enum: route.associationAliases,
                        },
                        description: 'The name of the association',
                    });
                },
                addAssociationFKPathParam: function () {
                    if (isParam('path', 'childid'))
                        return; // pre-defined/pre-configured --> skip
                    pathEntry.parameters.push({
                        in: 'path',
                        name: 'childid',
                        required: true,
                        schema: {
                            oneOf: exports.generateModelAssociationFKAttributeSchemas(route.model, route.associationAliases, models),
                        },
                        description: 'The desired target association record\'s foreign key value'
                    });
                },
                addAssociationResultOfArray: function () {
                    var _a;
                    var associations = ((_a = route.model) === null || _a === void 0 ? void 0 : _a.associations) || [];
                    var models = (route.associationAliases || []).map(function (a) {
                        var assoc = associations.find(function (_assoc) { return _assoc.alias == a; });
                        return assoc ? (assoc.collection || assoc.model) : a;
                    });
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: {
                                    type: 'array',
                                    // items: { type: 'any' },
                                    items: {
                                        oneOf: models.map(function (model) {
                                            return { '$ref': '#/components/schemas/' + model };
                                        }),
                                    },
                                },
                            },
                        },
                    });
                },
                addResultOfModel: function () {
                    assign_1.default(pathEntry.responses['200'], {
                        description: subst_1(template_1.resultDescription),
                        content: {
                            'application/json': {
                                schema: { '$ref': '#/components/schemas/' + route.model.identity },
                            },
                        },
                    });
                },
                addResultNotFound: function () {
                    assign_1.default(pathEntry.responses['404'], {
                        description: subst_1(template_1.notFoundDescription || 'Not found'),
                    });
                },
                addResultValidationError: function () {
                    assign_1.default(pathEntry.responses['400'], {
                        description: subst_1('Validation errors; details in JSON response'),
                    });
                },
                addFksBodyParam: function () {
                    if (route.actionType === 'shortcutBlueprint') {
                        exports.generateModelAssociationFKAttributeParameters(route.model, route.associationAliases, models).map(function (p) {
                            if (!route.associationAliases || route.associationAliases.indexOf(p.name) < 0)
                                return;
                            if (isParam('query', p.name))
                                return;
                            pathEntry.parameters.push(p);
                        });
                    }
                    else {
                        if (pathEntry.requestBody)
                            return;
                        pathEntry.requestBody = {
                            description: 'The primary key values (usually IDs) of the child records to use as the new members of this collection',
                            required: true,
                            content: {
                                'application/json': {
                                    schema: {
                                        type: 'array',
                                        items: {
                                            oneOf: exports.generateModelAssociationFKAttributeSchemas(route.model, route.associationAliases, models),
                                        }
                                    },
                                },
                            },
                        };
                    }
                },
                addShortCutBlueprintRouteNote: function () {
                    if (route.actionType !== 'shortcutBlueprint') {
                        return;
                    }
                    pathEntry.summary += ' *';
                    if (route.action === 'replace') {
                        pathEntry.description += "\n\nOnly one of the query parameters, that matches the **association** path parameter, should be specified.";
                    }
                    pathEntry.description += "\n\n(\\*) Note that this is a"
                        + " [Sails blueprint shortcut route](https://sailsjs.com/documentation/concepts/blueprints/blueprint-routes#?shortcut-blueprint-routes)"
                        + " (recommended for **development-mode only**)";
                },
            };
            // apply changes for blueprint action
            (template_1.modifiers || []).map(function (modifier) {
                if (isFunction_1.default(modifier))
                    modifier(template_1, route, pathEntry, tags, components); // custom modifier
                else
                    modifiers_1[modifier](); // standard modifier
            });
            defaults_1.default(pathEntry.responses, {
                '500': {
                    description: 'Internal server error',
                }
            });
        }
        else { // otherwise action
            pathEntry = __assign({ parameters: [], responses: cloneDeep_1.default(defaultsValues.responses || {}) }, cloneDeep_1.default(lodash_1.omit(route.swagger || {}, 'exclude')));
            if (route.actionType === 'actions2') {
                var patternVariables_1 = route.variables || [];
                if ((_j = route.actions2Machine) === null || _j === void 0 ? void 0 : _j.inputs) {
                    forEach_1.default(route.actions2Machine.inputs, function (value, key) {
                        var _a, _b;
                        if (((_b = (_a = value.meta) === null || _a === void 0 ? void 0 : _a.swagger) === null || _b === void 0 ? void 0 : _b.exclude) === true) {
                            return;
                        }
                        var _in = patternVariables_1.indexOf(key) >= 0 ? 'path' : 'query';
                        var isExisting = pathEntry.parameters.find(function (p) { return 'in' in p && p.in === _in && p.name === key; });
                        if (isExisting) {
                            return;
                        }
                        var description = value.description, _attribute = __rest(value, ["description"]);
                        var attribute = __assign(__assign({}, lodash_1.omit(_attribute, utils_1.attributeValidations)), { validations: pick_1.default(_attribute, utils_1.attributeValidations) });
                        pathEntry.parameters.push({
                            in: _in,
                            name: key,
                            required: value.required || false,
                            schema: exports.generateAttributeSchema(attribute),
                            description: description
                        });
                    });
                }
                if ((_k = route.actions2Machine) === null || _k === void 0 ? void 0 : _k.exits) {
                    var exitResponses_1 = {};
                    // actions2 may specify more than one 'exit' per 'statusCode' --> use oneOf (and attempt to merge)
                    forEach_1.default(route.actions2Machine.exits, function (exit, exitName) {
                        var _a, _b;
                        if (((_b = (_a = exit.meta) === null || _a === void 0 ? void 0 : _a.swagger) === null || _b === void 0 ? void 0 : _b.exclude) === true) {
                            return;
                        }
                        var _c = type_formatter_1.actions2Responses[exitName] || type_formatter_1.actions2Responses.success, statusCode = _c.statusCode, description = _c.description;
                        statusCode = exit.statusCode || statusCode;
                        description = exit.description || description;
                        // XXX TODO review support for responseType, viewTemplatePath, outputExample
                        // XXX TODO use deriveJsonSwaggerTypeFromExample() for 'content' otherwise don't use
                        if (exitResponses_1[statusCode]) {
                            // this statusCode already exists --> will contain 'oneOf' so add
                            var arr = exitResponses_1[statusCode].content['application/json'].schema.oneOf;
                            arr.push({ type: 'object', description: description });
                        }
                        else if (pathEntry.responses[statusCode]) {
                            // if not exists, check for response defined in source swagger and merge/massage to suit 'application/json' oneOf
                            var r = cloneDeep_1.default(pathEntry.responses[statusCode]);
                            if (!r.content)
                                r.content = {};
                            if (!r.content['application/json'])
                                r.content['application/json'] = {};
                            if (r.content['application/json'].schema) {
                                if (!r.content['application/json'].schema.oneOf) {
                                    var s = r.content['application/json'].schema;
                                    r.content['application/json'].schema = { oneOf: [s] };
                                }
                            }
                            else {
                                r.content['application/json'].schema = { oneOf: [] };
                            }
                            // now add this exit to the merged/massaged response
                            var arr = r.content['application/json'].schema.oneOf;
                            arr.push({ type: 'object', description: description });
                            exitResponses_1[statusCode] = r;
                        }
                        else {
                            // dne, so add
                            exitResponses_1[statusCode] = {
                                description: description,
                                content: {
                                    'application/json': {
                                        schema: { oneOf: [{ type: 'object', description: description }] }
                                    }
                                }
                            };
                        }
                    });
                    // remove oneOf for single entries
                    forEach_1.default(exitResponses_1, function (resp) {
                        var _a, _b;
                        if ((_b = (_a = resp.content) === null || _a === void 0 ? void 0 : _a['application/json'].schema) === null || _b === void 0 ? void 0 : _b.oneOf) {
                            var arr = resp.content['application/json'].schema.oneOf;
                            if (arr.length === 1) {
                                resp.content['application/json'].schema = arr[0];
                            }
                        }
                    });
                    pathEntry.responses = __assign(__assign({}, pathEntry.responses), exitResponses_1);
                    forEach_1.default(pathEntry.responses, function (resp, statusCode) {
                        var _a;
                        if (!resp.description)
                            resp.description = ((_a = exitResponses_1[statusCode]) === null || _a === void 0 ? void 0 : _a.description) || '-';
                    });
                }
                // actions2 summary and description
                defaults_1.default(pathEntry, {
                    summary: ((_l = route.actions2Machine) === null || _l === void 0 ? void 0 : _l.friendlyName) || route.path || '',
                    description: ((_m = route.actions2Machine) === null || _m === void 0 ? void 0 : _m.description) || undefined,
                    tags: ((_o = route.swagger) === null || _o === void 0 ? void 0 : _o.tags) || [((_p = route.actions2Machine) === null || _p === void 0 ? void 0 : _p.friendlyName) || route.defaultTagName],
                });
            } // of if(actions2)
            // final populate noting others above
            defaults_1.default(pathEntry, {
                summary: route.path || '',
                tags: ((_q = route.swagger) === null || _q === void 0 ? void 0 : _q.tags) || [route.defaultTagName],
            });
            // catch the case where defaultTagName not defined
            if (lodash_1.isEqual(pathEntry.tags, [undefined]))
                pathEntry.tags = [];
        }
        if (route.variables) {
            // now add patternVariables that don't already exist
            route.variables.map(function (v) {
                var existing = isParam('path', v);
                if (existing)
                    return;
                pathEntry.parameters.push({
                    in: 'path',
                    name: v,
                    required: true,
                    schema: { type: 'string' },
                    description: "Route pattern variable `" + v + "`",
                });
            });
        }
        if (pathEntry.tags) {
            pathEntry.tags.sort();
        }
        set_1.default(paths, [route.path, route.verb], pathEntry);
    });
    return paths;
};
exports.generateDefaultModelTags = function (models) {
    return lodash_1.map(models, function (model) {
        var _a, _b;
        var defaultDescription = "Sails blueprint actions for the **" + model.globalId + "** model";
        var tagDef = {
            name: model.globalId,
            description: ((_a = model.swagger.modelSchema) === null || _a === void 0 ? void 0 : _a.description) || defaultDescription,
        };
        if ((_b = model.swagger.modelSchema) === null || _b === void 0 ? void 0 : _b.externalDocs) {
            tagDef.externalDocs = __assign({}, model.swagger.modelSchema.externalDocs);
        }
        return tagDef;
    });
};
